use log::info;
use std::{
    env::{self, VarError},
    fs, io,
    path::Path,
    path::PathBuf,
    str::FromStr,
};
use tss_esapi::{tcti_ldr::TabrmdConfig, Context};

use btlib::{
    blocktree::{Blocktree, ModeAuthorizer},
    crypto::{
        tpm::{TpmCredStore, TpmCreds},
        CredStore,
    },
};

const PATH_ENV: &str = "BT_DIR";
const TABRMD_ENV: &str = "BT_TABRMD";
const TABRMD_CONFIG: &str = "bus_type=session,bus_name=com.intel.tss2.Tabrmd";
const STATE_FILE: &str = "tpm_state";
const BLOCK_DIR: &str = "blocks";

fn btdir() -> PathBuf {
    env::var(PATH_ENV).map(PathBuf::from).unwrap_or_else(|err| {
        if let VarError::NotPresent = err {
            let mut path_buf: PathBuf = env::var("HOME").unwrap().into();
            path_buf.push(".config");
            path_buf.push("blocktree");
            path_buf
        } else {
            panic!("invalid value of the '{PATH_ENV}' environment variable");
        }
    })
}

fn tabrmd_config() -> String {
    env::var(TABRMD_ENV).unwrap_or_else(|err| {
        if let VarError::NotPresent = err {
            TABRMD_CONFIG.to_string()
        } else {
            panic!("invalid value of the '{TABRMD_ENV}' environment variable");
        }
    })
}

fn make_blocktree(btdir: &Path, creds: TpmCreds) -> Blocktree<TpmCreds, ModeAuthorizer> {
    let block_dir = btdir.join(BLOCK_DIR);
    let exists = match fs::read_dir(&block_dir) {
        Ok(_) => true,
        Err(err) => {
            if let io::ErrorKind::NotFound = err.kind() {
                false
            } else {
                panic!("unexpected error occurred when checking if block dir exists: {err}");
            }
        }
    };
    if exists {
        Blocktree::new_existing(block_dir, creds, ModeAuthorizer {}).unwrap()
    } else {
        Blocktree::new_empty(block_dir, 0, creds, ModeAuthorizer {}).unwrap()
    }
}

fn main() {
    let btdir = btdir();
    info!("using btdir= '{}'", btdir.display());
    fs::create_dir_all(&btdir).unwrap();
    let tabrmd_config = tabrmd_config();
    info!("using tabrmd_config = '{tabrmd_config}'");
    let context =
        Context::new_with_tabrmd(TabrmdConfig::from_str(&tabrmd_config).unwrap()).unwrap();
    let cred_store = TpmCredStore::new(context, btdir.join(STATE_FILE)).unwrap();
    let node_creds = cred_store.node_creds().unwrap();
    let _bt = make_blocktree(&btdir, node_creds);
}
