use btlib::{error::BoxInIoErr, Result};
use btserde::{from_slice, read_from};
use bytes::BytesMut;
use futures::Future;
use serde::Deserialize;
use tokio::io::{AsyncRead, AsyncReadExt};

pub struct CallbackFramed<I> {
    io: I,
    buffer: BytesMut,
}

impl<I> CallbackFramed<I> {
    const INIT_CAPACITY: usize = 4096;

    pub fn new(inner: I) -> Self {
        Self {
            io: inner,
            buffer: BytesMut::with_capacity(Self::INIT_CAPACITY),
        }
    }

    async fn decode<'de, F: 'de + DeserCallback>(
        mut slice: &'de [u8],
        callback: &'de F,
    ) -> Result<DecodeStatus<F::Return>> {
        let payload_len: u64 = match read_from(&mut slice) {
            Ok(payload_len) => payload_len,
            Err(err) => {
                return match err {
                    btserde::Error::Eof => Ok(DecodeStatus::None),
                    btserde::Error::Io(ref io_err) => match io_err.kind() {
                        std::io::ErrorKind::UnexpectedEof => Ok(DecodeStatus::None),
                        _ => Err(err.into()),
                    },
                    _ => Err(err.into()),
                }
            }
        };
        let payload_len: usize = payload_len.try_into().box_err()?;
        if slice.len() < payload_len {
            return Ok(DecodeStatus::Reserve(payload_len - slice.len()));
        }
        let msg: F::Arg<'de> = from_slice(slice)?;
        let returned = callback.call(msg).await;
        Ok(DecodeStatus::Some {
            returned,
            consumed: std::mem::size_of::<u64>() + payload_len,
        })
    }
}

macro_rules! attempt {
    ($result:expr) => {
        match $result {
            Ok(value) => value,
            Err(err) => return Some(Err(err.into())),
        }
    };
}

impl<S: AsyncRead + Unpin> CallbackFramed<S> {
    pub async fn next<F: DeserCallback>(&mut self, callback: F) -> Option<Result<F::Return>> {
        loop {
            if self.buffer.capacity() - self.buffer.len() == 0 {
                // If there is no space left in the buffer we reserve additional bytes to ensure
                // read_buf doesn't return 0 unless we're at EOF.
                self.buffer.reserve(1);
            }
            let read_ct = attempt!(self.io.read_buf(&mut self.buffer).await);
            if 0 == read_ct {
                return None;
            }
            match attempt!(Self::decode(&self.buffer[..read_ct], &callback).await) {
                DecodeStatus::None => continue,
                DecodeStatus::Reserve(count) => {
                    self.buffer.reserve(count);
                    continue;
                }
                DecodeStatus::Some { returned, consumed } => {
                    let _ = self.buffer.split_to(consumed);
                    return Some(Ok(returned));
                }
            }
        }
    }
}

enum DecodeStatus<R> {
    None,
    Reserve(usize),
    Some { returned: R, consumed: usize },
}

pub trait DeserCallback: Clone {
    type Arg<'de>: Deserialize<'de> + Send
    where
        Self: 'de;
    type Return;
    type CallFut<'s>: Future<Output = Self::Return> + Send
    where
        Self: 's;
    fn call<'de>(&'de self, arg: Self::Arg<'de>) -> Self::CallFut<'de>;
}

impl<F: DeserCallback> DeserCallback for &F {
    type Arg<'de> = F::Arg<'de> where Self: 'de;
    type Return = F::Return;
    type CallFut<'f> = F::CallFut<'f> where Self: 'f;
    fn call<'de>(&'de self, arg: Self::Arg<'de>) -> Self::CallFut<'de> {
        (*self).call(arg)
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    use crate::MsgEncoder;

    use futures::{future::Ready, SinkExt};
    use serde::Serialize;
    use std::io::{Cursor, Seek};
    use tokio_util::codec::FramedWrite;

    #[derive(Serialize, Deserialize)]
    struct Msg<'a>(&'a [u8]);

    #[tokio::test]
    async fn read_single_message() {
        macro_rules! test_data {
            () => {
                b"fulcrum"
            };
        }

        #[derive(Clone)]
        struct TestCb;

        impl DeserCallback for TestCb {
            type Arg<'de> = Msg<'de> where Self: 'de;
            type Return = bool;
            type CallFut<'f> = Ready<bool>;

            fn call<'de>(&'de self, arg: Self::Arg<'de>) -> Self::CallFut<'de> {
                futures::future::ready(arg.0 == test_data!())
            }
        }

        let mut write = FramedWrite::new(Cursor::new(Vec::<u8>::new()), MsgEncoder);
        write.send(Msg(test_data!())).await.unwrap();
        let mut io = write.into_inner();
        io.rewind().unwrap();
        let mut read = CallbackFramed::new(io);

        let matched = read.next(TestCb).await.unwrap().unwrap();

        assert!(matched)
    }
}
