//! Code which enables sending messages between processes in the blocktree system.

mod tls;
use tls::*;
mod callback_framed;
use callback_framed::{CallbackFramed, DeserCallback};

use btlib::{bterr, crypto::Creds, error::BoxInIoErr, BlockPath, Result, Writecap};
use btserde::{read_from, write_to};
use bytes::{BufMut, BytesMut};
use core::{future::Future, marker::Send, ops::DerefMut, pin::Pin};
use futures::{sink::Send as SendFut, SinkExt, StreamExt};
use log::error;
use quinn::{Connection, Endpoint, RecvStream, SendStream};
use serde::{de::DeserializeOwned, Deserialize, Serialize};
use std::{
    any::Any,
    collections::hash_map::DefaultHasher,
    hash::{Hash, Hasher},
    marker::PhantomData,
    net::{IpAddr, Ipv6Addr, SocketAddr},
    sync::Arc,
};
use tokio::{
    select,
    sync::{broadcast, Mutex},
};
use tokio_util::codec::{Decoder, Encoder, Framed, FramedParts, FramedRead, FramedWrite};

/// Returns a [Receiver] bound to the given [IpAddr] which receives messages at the bind path of
/// the given [Writecap] of the given credentials. The returned type can be used to make
/// [Transmitter]s for any path.
pub fn receiver<C: 'static + Creds + Send + Sync, F: 'static + MsgCallback>(
    ip_addr: IpAddr,
    creds: Arc<C>,
    callback: F,
) -> Result<impl Receiver> {
    let writecap = creds.writecap().ok_or(btlib::BlockError::MissingWritecap)?;
    let addr = Arc::new(BlockAddr::new(ip_addr, Arc::new(writecap.bind_path())));
    QuicReceiver::new(addr, Arc::new(CertResolver::new(creds)?), callback)
}

pub async fn transmitter<C: 'static + Creds + Send + Sync>(
    addr: Arc<BlockAddr>,
    creds: Arc<C>,
) -> Result<impl Transmitter> {
    let resolver = Arc::new(CertResolver::new(creds)?);
    let endpoint = Endpoint::client(SocketAddr::new(IpAddr::V6(Ipv6Addr::UNSPECIFIED), 0))?;
    QuicTransmitter::from_endpoint(endpoint, addr, resolver).await
}

pub trait MsgCallback: Clone + Send + Sync + Unpin {
    type Arg<'de>: CallMsg<'de>
    where
        Self: 'de;
    type Return;
    type CallFut<'de>: Future<Output = Self::Return> + Send
    where
        Self: 'de;
    fn call<'de>(&'de self, arg: MsgReceived<Self::Arg<'de>>) -> Self::CallFut<'de>;
}

impl<T: MsgCallback> MsgCallback for &T {
    type Arg<'de> = T::Arg<'de> where Self: 'de;
    type Return = T::Return;
    type CallFut<'de> = T::CallFut<'de> where Self: 'de;
    fn call<'de>(&'de self, arg: MsgReceived<Self::Arg<'de>>) -> Self::CallFut<'de> {
        (*self).call(arg)
    }
}

/// Trait for messages which can be transmitted using the call method.
pub trait CallMsg<'de>: Serialize + Deserialize<'de> + Send + Sync {
    type Reply: Serialize + DeserializeOwned + Send;
}

#[derive(Serialize, Deserialize)]
pub enum NoReply {}

/// Trait for messages which can be transmitted using the send method.
/// Types which implement this trait should specify [NoReply] as their reply type.
pub trait SendMsg<'de>: CallMsg<'de> {}

/// An address which identifies a block on the network. An instance of this struct can be
/// used to get a socket address for the block this address refers to.
#[derive(PartialEq, Eq, Hash, Clone, Debug)]
pub struct BlockAddr {
    ip_addr: IpAddr,
    path: Arc<BlockPath>,
}

impl BlockAddr {
    pub fn new(ip_addr: IpAddr, path: Arc<BlockPath>) -> Self {
        Self { ip_addr, path }
    }

    pub fn ip_addr(&self) -> IpAddr {
        self.ip_addr
    }

    pub fn path(&self) -> &BlockPath {
        self.path.as_ref()
    }

    fn port(&self) -> u16 {
        // TODO: We should probably choose a stable hasher, as the standard library devs could
        // change `DefaultHasher` at any time.
        let mut hasher = DefaultHasher::new();
        self.path.hash(&mut hasher);
        let hash = hasher.finish();
        // We compute a port in the dynamic range [49152, 65535] as defined by RFC 6335.
        const NUM_RES_PORTS: u16 = 49153;
        const PORTS_AVAIL: u64 = (u16::MAX - NUM_RES_PORTS) as u64;
        NUM_RES_PORTS + (hash % PORTS_AVAIL) as u16
    }

    /// Returns the socket address of the block this instance refers to.
    pub fn socket_addr(&self) -> SocketAddr {
        SocketAddr::new(self.ip_addr, self.port())
    }
}

/// Indicates whether a message was sent using `call` or `send`.
#[derive(Debug, Serialize, Deserialize, PartialEq, Eq, Hash, Clone)]
enum MsgKind {
    /// This message expects exactly one reply.
    Call,
    /// This message expects exactly zero replies.
    Send,
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq, Hash, Clone)]
struct Envelope<T> {
    kind: MsgKind,
    msg: T,
}

impl<T> Envelope<T> {
    fn send(msg: T) -> Self {
        Self {
            msg,
            kind: MsgKind::Send,
        }
    }

    fn call(msg: T) -> Self {
        Self {
            msg,
            kind: MsgKind::Call,
        }
    }

    fn msg(&self) -> &T {
        &self.msg
    }
}

/// A message tagged with the block path that it was sent from.
pub struct MsgReceived<T> {
    from: Arc<BlockPath>,
    msg: Envelope<T>,
    replier: Option<Replier>,
}

impl<T> MsgReceived<T> {
    fn new(from: Arc<BlockPath>, msg: Envelope<T>, replier: Option<Replier>) -> Self {
        Self { from, msg, replier }
    }

    /// The path from which this message was received.
    pub fn from(&self) -> &Arc<BlockPath> {
        &self.from
    }

    /// Payload contained in this message.
    pub fn body(&self) -> &T {
        self.msg.msg()
    }

    /// Returns true if and only if this messages needs to be replied to.
    pub fn needs_reply(&self) -> bool {
        self.replier.is_some()
    }
}

impl<'de, T: CallMsg<'de>> MsgReceived<T> {
    /// Returns a type which can be used to reply to this message, if this message requires a
    /// reply and it has not yet been sent.
    pub fn take_replier(&mut self) -> Option<Replier> {
        self.replier.take()
    }
}

/// Trait for receiving messages and creating [Transmitter]s.
pub trait Receiver {
    /// The address at which messages will be received.
    fn addr(&self) -> &Arc<BlockAddr>;

    type Transmitter: Transmitter + Send;
    type TransmitterFut<'a>: 'a + Future<Output = Result<Self::Transmitter>> + Send
    where
        Self: 'a;

    /// Creates a [Transmitter] which is connected to the given address.
    fn transmitter(&self, addr: Arc<BlockAddr>) -> Self::TransmitterFut<'_>;
}

/// A type which can be used to transmit messages.
pub trait Transmitter {
    type SendFut<'s, T>: 's + Future<Output = Result<()>> + Send
    where
        Self: 's,
        T: 's + Serialize + Send;

    /// Transmit a message to the connected [Receiver] without waiting for a reply.
    fn send<'de, T: 'de + SendMsg<'de>>(&'de mut self, msg: T) -> Self::SendFut<'de, T>;

    type CallFut<'s, 'de, T>: 's + Future<Output = Result<T::Reply>> + Send
    where
        Self: 's,
        T: 's + CallMsg<'de>,
        T::Reply: 's;

    /// Transmit a message to the connected [Receiver] and wait for a reply.
    fn call<'s, 'de, T>(&'s mut self, msg: T) -> Self::CallFut<'s, 'de, T>
    where
        T: 's + CallMsg<'de>,
        T::Reply: 's;

    type FinishFut: Future<Output = Result<()>> + Send;

    /// Finish any ongoing transmissions and close the connection to the [Receiver].
    fn finish(self) -> Self::FinishFut;

    /// Returns the address that this instance is transmitting to.
    fn addr(&self) -> &Arc<BlockAddr>;
}

/// Encodes messages using [btserde].
struct MsgEncoder;

impl MsgEncoder {
    fn new() -> Self {
        Self
    }
}

impl<T: Serialize> Encoder<T> for MsgEncoder {
    type Error = btlib::Error;

    fn encode(&mut self, item: T, dst: &mut BytesMut) -> Result<()> {
        const U64_LEN: usize = std::mem::size_of::<u64>();
        let payload = dst.split_off(U64_LEN);
        let mut writer = payload.writer();
        write_to(&item, &mut writer)?;
        let payload = writer.into_inner();
        let payload_len = payload.len() as u64;
        let mut writer = dst.writer();
        write_to(&payload_len, &mut writer)?;
        let dst = writer.into_inner();
        dst.unsplit(payload);
        Ok(())
    }
}

/// Decodes messages using [btserde].
struct MsgDecoder<T>(PhantomData<T>);

impl<T> MsgDecoder<T> {
    fn new() -> Self {
        Self(PhantomData)
    }
}

impl<T: DeserializeOwned> Decoder for MsgDecoder<T> {
    type Item = T;
    type Error = btlib::Error;

    fn decode(&mut self, src: &mut BytesMut) -> Result<Option<Self::Item>> {
        let mut slice: &[u8] = src.as_ref();
        let payload_len: u64 = match read_from(&mut slice) {
            Ok(payload_len) => payload_len,
            Err(err) => {
                return match err {
                    btserde::Error::Eof => Ok(None),
                    btserde::Error::Io(ref io_err) => match io_err.kind() {
                        std::io::ErrorKind::UnexpectedEof => Ok(None),
                        _ => Err(err.into()),
                    },
                    _ => Err(err.into()),
                }
            }
        };
        let payload_len: usize = payload_len.try_into().box_err()?;
        if slice.len() < payload_len {
            src.reserve(payload_len - slice.len());
            return Ok(None);
        }
        let msg = read_from(&mut slice)?;
        // Consume all the bytes that have been read out of the buffer.
        let _ = src.split_to(std::mem::size_of::<u64>() + payload_len);
        Ok(Some(msg))
    }
}

type SharedFrameParts = Arc<Mutex<Option<FramedParts<SendStream, MsgEncoder>>>>;

#[derive(Clone)]
pub struct Replier {
    parts: SharedFrameParts,
}

impl Replier {
    fn new(send_stream: SendStream) -> Self {
        let parts = FramedParts::new::<()>(send_stream, MsgEncoder::new());
        let parts = Arc::new(Mutex::new(Some(parts)));
        Self { parts }
    }

    pub async fn reply<T: Serialize + Send>(self, reply: T) -> Result<()> {
        let parts = self.parts;
        let mut guard = parts.lock().await;
        // We must ensure the parts are put back before we leave this block.
        let parts = guard.take().unwrap();
        let mut stream = Framed::from_parts(parts);
        let result = stream.send(reply).await;
        *guard = Some(stream.into_parts());
        result
    }
}

#[derive(Clone)]
struct MsgRecvdCallback<F> {
    path: Arc<BlockPath>,
    replier: Replier,
    inner: F,
}

impl<F: MsgCallback> MsgRecvdCallback<F> {
    fn new(path: Arc<BlockPath>, replier: Replier, inner: F) -> Self {
        Self {
            path,
            replier,
            inner,
        }
    }
}

impl<F: MsgCallback> DeserCallback for MsgRecvdCallback<F> {
    type Arg<'de> = Envelope<F::Arg<'de>> where Self: 'de;
    type Return = F::Return;
    type CallFut<'s> = F::CallFut<'s> where F: 's;
    fn call<'de>(&'de self, arg: Envelope<F::Arg<'de>>) -> Self::CallFut<'de> {
        let replier = match arg.kind {
            MsgKind::Call => Some(self.replier.clone()),
            MsgKind::Send => None,
        };
        self.inner
            .call(MsgReceived::new(self.path.clone(), arg, replier))
    }
}

macro_rules! handle_err {
    ($result:expr, $on_err:expr, $control_flow:expr) => {
        match $result {
            Ok(inner) => inner,
            Err(err) => {
                $on_err(err);
                $control_flow;
            }
        }
    };
}

/// Unwraps the given result, or if the result is an error, returns from the enclosing function.
macro_rules! unwrap_or_return {
    ($result:expr, $on_err:expr) => {
        handle_err!($result, $on_err, return)
    };
    ($result:expr) => {
        unwrap_or_return!($result, |err| error!("{err}"))
    };
}

/// Unwraps the given result, or if the result is an error, continues the enclosing loop.
macro_rules! unwrap_or_continue {
    ($result:expr, $on_err:expr) => {
        handle_err!($result, $on_err, continue)
    };
    ($result:expr) => {
        unwrap_or_continue!($result, |err| error!("{err}"))
    };
}

/// Awaits its first argument, unless interrupted by its second argument, in which case the
/// enclosing function returns. The second argument needs to be cancel safe, but the first
/// need not be if it is discarded when the enclosing function returns (because losing messages
/// from the first argument doesn't matter in this case).
macro_rules! await_or_stop {
    ($future:expr, $stop_fut:expr) => {
        select! {
            Some(connecting) = $future => connecting,
            _ = $stop_fut => return,
        }
    };
}

struct QuicReceiver {
    recv_addr: Arc<BlockAddr>,
    stop_tx: broadcast::Sender<()>,
    endpoint: Endpoint,
    resolver: Arc<CertResolver>,
}

impl QuicReceiver {
    fn new<F: 'static + MsgCallback>(
        recv_addr: Arc<BlockAddr>,
        resolver: Arc<CertResolver>,
        callback: F,
    ) -> Result<Self> {
        let socket_addr = recv_addr.socket_addr();
        let endpoint = Endpoint::server(server_config(resolver.clone())?, socket_addr)?;
        let (stop_tx, stop_rx) = broadcast::channel(1);
        tokio::spawn(Self::server_loop(endpoint.clone(), callback, stop_rx));
        Ok(Self {
            recv_addr,
            stop_tx,
            endpoint,
            resolver,
        })
    }

    async fn server_loop<F: 'static + MsgCallback>(
        endpoint: Endpoint,
        callback: F,
        mut stop_rx: broadcast::Receiver<()>,
    ) {
        loop {
            let connecting = await_or_stop!(endpoint.accept(), stop_rx.recv());
            let connection = unwrap_or_continue!(connecting.await, |err| error!(
                "error accepting QUIC connection: {err}"
            ));
            tokio::spawn(Self::handle_connection(
                connection,
                callback.clone(),
                stop_rx.resubscribe(),
            ));
        }
    }

    async fn handle_connection<F: MsgCallback>(
        connection: Connection,
        callback: F,
        mut stop_rx: broadcast::Receiver<()>,
    ) {
        let client_path = unwrap_or_return!(
            Self::client_path(connection.peer_identity()),
            |err| error!("failed to get client path from peer identity: {err}")
        );
        let (send_stream, recv_stream) = unwrap_or_return!(
            connection.accept_bi().await,
            |err| error!("error accepting receive stream: {err}")
        );
        let replier = Replier::new(send_stream);
        let callback = MsgRecvdCallback::new(client_path, replier, callback);
        let mut msg_stream = CallbackFramed::new(recv_stream);
        loop {
            let decode_result = await_or_stop!(msg_stream.next(callback.clone()), stop_rx.recv());
            if let Err(ref err) = decode_result {
                error!("msg_stream produced an error: {err}");
            }
        }
    }

    /// Returns the path the client is bound to.
    fn client_path(peer_identity: Option<Box<dyn Any>>) -> Result<Arc<BlockPath>> {
        let peer_identity =
            peer_identity.ok_or_else(|| bterr!("connection did not contain a peer identity"))?;
        let client_certs = peer_identity
            .downcast::<Vec<rustls::Certificate>>()
            .map_err(|_| bterr!("failed to downcast peer_identity to certificate chain"))?;
        let first = client_certs
            .first()
            .ok_or_else(|| bterr!("no certificates were presented by the client"))?;
        let (writecap, ..) = Writecap::from_cert_chain(first, &client_certs[1..])?;
        Ok(Arc::new(writecap.bind_path()))
    }
}

impl Drop for QuicReceiver {
    fn drop(&mut self) {
        // This result will be a failure if the tasks have already returned, which is not a
        // problem.
        let _ = self.stop_tx.send(());
    }
}

impl Receiver for QuicReceiver {
    fn addr(&self) -> &Arc<BlockAddr> {
        &self.recv_addr
    }

    type Transmitter = QuicTransmitter;
    type TransmitterFut<'a> = Pin<Box<dyn 'a + Future<Output = Result<QuicTransmitter>> + Send>>;

    fn transmitter(&self, addr: Arc<BlockAddr>) -> Self::TransmitterFut<'_> {
        Box::pin(async {
            QuicTransmitter::from_endpoint(self.endpoint.clone(), addr, self.resolver.clone()).await
        })
    }
}

struct QuicTransmitter {
    addr: Arc<BlockAddr>,
    sink: FramedWrite<SendStream, MsgEncoder>,
    recv_stream: Mutex<RecvStream>,
}

impl QuicTransmitter {
    async fn from_endpoint(
        endpoint: Endpoint,
        addr: Arc<BlockAddr>,
        resolver: Arc<CertResolver>,
    ) -> Result<Self> {
        let socket_addr = addr.socket_addr();
        let connecting = endpoint.connect_with(
            client_config(addr.path.clone(), resolver)?,
            socket_addr,
            // The ServerCertVerifier ensures we connect to the correct path.
            "UNIMPORTANT",
        )?;
        let connection = connecting.await?;
        let (send_stream, recv_stream) = connection.open_bi().await?;
        let sink = FramedWrite::new(send_stream, MsgEncoder::new());
        Ok(Self {
            addr,
            sink,
            recv_stream: Mutex::new(recv_stream),
        })
    }
}

/// TODO: Once the "Permit impl Trait in type aliases"
/// https://github.com/rust-lang/rust/issues/63063
/// feature lands the future types in this implementation should be rewritten to
/// use it.
impl Transmitter for QuicTransmitter {
    fn addr(&self) -> &Arc<BlockAddr> {
        &self.addr
    }

    type SendFut<'s, T> = SendFut<'s, FramedWrite<SendStream, MsgEncoder>, Envelope<T>>
        where T: 's + Serialize + Send;

    fn send<'de, T: 'de + SendMsg<'de>>(&'de mut self, msg: T) -> Self::SendFut<'de, T> {
        self.sink.send(Envelope::send(msg))
    }

    type FinishFut = Pin<Box<dyn Future<Output = Result<()>> + Send>>;

    fn finish(mut self) -> Self::FinishFut {
        Box::pin(async move {
            let steam: &mut SendStream = self.sink.get_mut();
            steam.finish().await.map_err(|err| bterr!(err))
        })
    }

    type CallFut<'s, 'de, T> = Pin<Box<dyn 's + Future<Output = Result<T::Reply>> + Send>>
    where
        T: 's + CallMsg<'de>,
        T::Reply: 's;

    fn call<'s, 'de, T>(&'s mut self, msg: T) -> Self::CallFut<'s, 'de, T>
    where
        T: 's + CallMsg<'de>,
        T::Reply: 's,
    {
        Box::pin(async move {
            self.sink.send(Envelope::call(msg)).await?;
            let mut guard = self.recv_stream.lock().await;
            let mut source = FramedRead::new(guard.deref_mut(), MsgDecoder::<T::Reply>::new());
            source
                .next()
                .await
                .ok_or_else(|| bterr!("server hung up before sending reply"))?
        })
    }
}
